package remote;

import java.util.Collection;
import java.util.Iterator;
import java.util.Random;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.framework.recipes.locks.InterProcessMutex;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.data.Stat;

import tpool.ConcurrentRingBuffer;

public class RemoteFreeList<T extends Freeable<V, U>, V, U> {
	
	
	private static final int SLEEP_TIME = 3*1000;
	private static final int CACHE_TIME = 5*1000;
	private int CACHE_TRIES = 4;
	
	private LockManager listManager;
	private LockManager objManager;
	private RemoteMap objects;
	private CuratorFramework client;
	private ObjectFactory<T, V> generator;
	
	private String base;
	private String baseSlash;
	private Random random = new Random();
	
	public AtomicInteger cacheTries;
	public AtomicInteger cacheHits;
	
	private ConcurrentRingBuffer<T> cache;

	
	public RemoteFreeList(String basePath, CuratorFramework client, ObjectFactory<T,V> gen, LockManager list,
			LockManager objs, RemoteMap map, int cacheSize) throws Exception
	{
		cacheTries = new AtomicInteger(0);
		cacheHits = new AtomicInteger(0);
		
		cache = new ConcurrentRingBuffer<T>(cacheSize);
		CACHE_TRIES = Math.min(CACHE_TRIES,cacheSize);
		random.setSeed(System.nanoTime());
		this.client = client;
		base = basePath;
		baseSlash = base + "/";
		this.generator = gen;
		try
		{
			client.create().creatingParentsIfNeeded().forPath(basePath);
		}
		catch(KeeperException.NodeExistsException e){}
		objManager = objs;
		listManager = list;
		objects = map;
		new Cleaner().start();
		
	}
	
	private boolean upToDate(T item) throws Exception
	{
		String key = item.getKey();
		InterProcessMutex lock = objManager.tryLock(key);
		if(lock == null) return false;
		Stat s = objects.getStat(key);
		
		if(s.getVersion() == item.getVersion())
		{
			item.setLock(lock);
			return true;
		}
		objManager.releaseLock(lock);
		return false;
	}
	
	private T tryCache(V restriction, U keyRestriction) throws Exception
	{
		long time = System.currentTimeMillis();
		cacheTries.incrementAndGet();
		for(int i = 0; i < CACHE_TRIES; i++)
		{
			T item = cache.tryTake();
			if(item == null)
			{
				return null;
			}
			if(item.free(restriction,keyRestriction))
			{
				if(time-item.getTime() > CACHE_TIME)
				{
					continue;
				}
				if(upToDate(item))
				{
					cacheHits.incrementAndGet();
					return item;
				}
			}
			cache.put(item);
		}
		return null;
	}
	
	private void putCache(T item)
	{
		item.setTime(System.currentTimeMillis());
		cache.tryPut(item);
	}
	
	public T handleEmptyCase(V res) throws Exception
	{
		String key;
		do
		{
			try
			{
				long value = random.nextLong();
				key = ""+value;
				client.create().forPath(baseSlash + key);
				T obj = generator.generate(res);
				obj.setRemoteGroup(key);
				obj.setVersion(-1);
				return obj;
			}
			catch(KeeperException.NodeExistsException e)
			{
				
			}
		}while(true);
	}
	
	private T handleEmptyCase(V res, String owner, InterProcessMutex m) throws Exception
	{
		listManager.releaseLock(m);
		T obj = generator.generate(res);
		obj.setRemoteGroup(owner);
		obj.setVersion(-1);
		return obj;
	}
	
	public T getFreeItem(V restriction, U keyRestriction) throws Exception
	{
		String key = "";
		T obj = tryCache(restriction, keyRestriction);
		if(obj != null)
		{
			return obj;
		}
		Collection<String> children = client.getChildren().forPath(base);
		if(children.size() <= 3)
		{
			return handleEmptyCase(restriction);
		}
		else
		{	
			InterProcessMutex m = null;
			int j = 0;
			while(m == null && j++ < 16)
			{
				int num = (int)(children.size()*Math.random());
				Iterator<String> t = children.iterator();
				for(int i = 0; i < num; i++)
				{
					t.next();
				}
				key = t.next();
				m = listManager.tryLock(key);
			}
			if(m == null)
			{
				return handleEmptyCase(restriction);
			}
			Collection<String> subChildren;
			try
			{
				subChildren = client.getChildren().forPath(baseSlash+key);
			}
			catch(KeeperException.NoNodeException e)
			{
				return handleEmptyCase(restriction, key, m);
			}
			if(subChildren.size() == 0)
			{
				return handleEmptyCase(restriction, key, m);
			}
			Iterator<String> sub = subChildren.iterator();
			String subkey;
			while(sub.hasNext())
			{
				subkey = sub.next();
				Stat stat = new Stat();
				byte[] data = objects.get(subkey, stat);
				obj = generator.generateBlank(restriction);
				InterProcessMutex lock = objManager.tryLock(subkey);
				if(lock == null) continue;
				obj.fromBytes(data);
				if(obj.free(restriction, keyRestriction))
				{
					obj.setLock(lock);
					obj.setVersion(stat.getVersion());
					obj.setRemoteGroup(key);
					obj.parseKey(subkey);
					if(obj.free(restriction, keyRestriction))
					{
						listManager.releaseLock(m);
						return obj;
					}
				}
				objManager.releaseLock(lock);
			}
			if(subChildren.size() < 16)
			{
				return handleEmptyCase(restriction,key,m);
			}
			else
			{
				listManager.releaseLock(m);
				return handleEmptyCase(restriction);
			}
		}
	}

	public void putFreeItem(T item) throws Exception
	{
		String key = item.getKey();
		String owner = item.getRemoteGroup();
		String base  = baseSlash + owner;
		String bases = baseSlash + owner + "/";
		InterProcessMutex m = listManager.getLock(owner);
		InterProcessMutex lock = item.getLock();
		try
		{
			if(!item.free(null,null))
			{
				try
				{
					client.delete().forPath(bases + key);
				}
				catch(KeeperException.NoNodeException e)
				{
					
				}
				Collection<String> subChildren = client.getChildren().forPath(base);
				if(subChildren.size() == 0)
				{
					try
					{
						client.delete().forPath(base);
					}
					catch(KeeperException.NoNodeException e)
					{
						
					}
				}
			}
			else
			{
				try
				{
					client.create().creatingParentsIfNeeded().forPath(bases + key);
				}
				catch(KeeperException.NodeExistsException e)
				{
					
				}
				this.putCache(item);
				listManager.releaseLock(m);
				m = null;
			}
		}
		finally
		{
			try
			{
				if(m != null)
				{
					listManager.releaseLock(m);
				}
				
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
			try
			{
				if(lock != null)
				{
					objManager.releaseLock(lock);
				}
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
		}
	}
	
	private class Cleaner extends Thread
	{
		public void run()
		{
			while(true)
			{
				try {
					Thread.sleep(SLEEP_TIME);
				} catch (InterruptedException e) {}
				T item = cache.tryTake();
				if(item == null ||
						System.currentTimeMillis() - item.getTime() > CACHE_TIME ||
						!item.free(null, null))
				{
					continue;
				}
				
				else
				{
					cache.tryPut(item);
				}
			}
		}
	}
}
