package remote;

import java.util.Collection;
import java.util.Iterator;

import objectmap.KeyData;
import objectmap.PutsTuple;

import org.apache.curator.framework.api.transaction.CuratorTransactionFinal;
import org.apache.curator.framework.recipes.locks.InterProcessMutex;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.data.Stat;

public class RemoteManager<K extends Versionable, V>
{
	private ManagerErrorHandler<K> handler;
	private LockManager lockManager;
	private LockManager clManager;
	private RemoteMap map;
	private String basePath;
	private String baseSlash;
	private String subPath;
	private String subSlash;
	private ObjectFactory<K,V> factory;
	private static final int SLEEP = 3 * 1000;
	
	
	public RemoteManager(String base, RemoteMap map,
			ManagerErrorHandler<K> handler, ObjectFactory<K, V> factory,
			int id, LockManager lm) throws Exception
	{
		basePath = base;
		baseSlash = base + "/";
		subPath = baseSlash + id;
		subSlash = subPath + "/";
		this.map = map;
		this.handler = handler;
		this.factory = factory;
		try
		{
			map.client.create().creatingParentsIfNeeded().forPath(subPath);
		}
		catch(KeeperException.NodeExistsException e)
		{
			
		}
		new Cleaner().start();
		lockManager = lm;
		clManager = null;
	}
	
	public String jobPath(String key, K item)
	{
		return subSlash+item.getUniqueKey(key);
	}
	
	public void registerJob(String key, K item) throws Exception
	{
		String job = subSlash+item.getUniqueKey(key);
		byte[] data = item.getBytes();
		map.client.create().creatingParentsIfNeeded().forPath(job, data);
	}
	
	public void deregisterJob(String key, K item) throws Exception
	{
		String job = subSlash+item.getUniqueKey(key);
		try
		{
			map.client.delete().forPath(job);
		}
		catch(KeeperException.NoNodeException e)
		{
			
		}
	}
	
	public void deregisterMass(PutsTuple[] pt, K[] kd,
			K[] okd, boolean[] first, int s, int e) throws Exception
	{
		CuratorTransactionFinal trans;
		int i = s;
		String key = pt[i].key;
		K item = kd[i];
		K oldItem = okd[i];
		
		String uk = item.getUniqueKey(key);
		String ouk = oldItem.getUniqueKey(key);
		String job = subSlash+uk;
		String oldJob = subSlash+ouk;
		int version = item.getVersion();
		item.incVersion();
		byte[] data = item.getBytes();
		byte[] oldData = oldItem.getBytes();
		String mapPath = map.putPath(key);
		
		if(first[i])
		{
			if(version != -1)
			{
				throw new Exception("Wrong Version DREG");
			}
			trans = map.client.inTransaction().create().forPath(mapPath,data).
			and().delete().forPath(job).and();
			
		}
		else
			trans = map.client.inTransaction().setData().
			withVersion(version).forPath(mapPath,data).
			and().delete().forPath(job).
			and().create().forPath(oldJob, oldData).and();
		
		for(i += 1; i < e; i++)
		{
			key = pt[i].key;
			item = kd[i];
			oldItem = okd[i];
			
			uk = item.getUniqueKey(key);
			ouk = oldItem.getUniqueKey(key);
			job = subSlash+uk;
			oldJob = subSlash+ouk;
			version = item.getVersion();
			item.incVersion();
			data = item.getBytes();
			oldData = oldItem.getBytes();
			mapPath = map.putPath(key);
			
			if(first[i])
			{
				if(version != -1)
				{
					throw new Exception("Wrong Version DREG");
				}
				trans.create().forPath(mapPath,data).
				and().delete().forPath(job).and();
				
			}
			else
				trans.setData().
				withVersion(version).forPath(mapPath,data).
				and().delete().forPath(job).
				and().create().forPath(oldJob, oldData).and();
			
		}
		trans.commit();
	}
	
	public void deregisterJobs(String key, K item, K oldItem,
			boolean first) throws Exception
	{
		String uk = item.getUniqueKey(key);
		String ouk = oldItem.getUniqueKey(key);
		String job = subSlash+uk;
		String oldJob = subSlash+ouk;
		int version = item.getVersion();
		item.incVersion();
		byte[] data = item.getBytes();
		byte[] oldData = oldItem.getBytes();
		String mapPath = map.putPath(key);
		if(first)
		{
			if(version != -1)
			{
				throw new Exception("Wrong Version DREG");
			}
			map.client.inTransaction().create().forPath(mapPath,data).
			and().delete().forPath(job).and().commit();
			
		}
		else
			map.client.inTransaction().setData().
			withVersion(version).forPath(mapPath,data).
			and().delete().forPath(job).
			and().create().forPath(oldJob, oldData).and().commit();
		
	}
	
	private void clean() throws Exception
	{
		while(true)
		{
			Collection<String> children = 
					map.client.getChildren().forPath(basePath);
			Iterator<String> it = children.iterator();
			while(it.hasNext())
			{
				String next = it.next();
				String subPath = baseSlash+next;
				Collection<String> subChildren;
				try
				{
					subChildren = map.client.getChildren().
							forPath(subPath);
				}
				catch(KeeperException.NoNodeException e)
				{
					continue;
				}
				Iterator<String> sit = subChildren.iterator();
				while(sit.hasNext()) {
					K item = factory.generate(null);
					String key = sit.next();
					
					String str = item.parseUniqueKey(key);
					InterProcessMutex m = lockManager.tryLock(str);
					
					if(m == null) continue;
					System.out.println(str);
					try
					{
						Stat s = map.getStat(str);
						byte[] data;
						try
						{
							data = map.client.getData().forPath(subPath+"/"+key);
						}
						catch(Exception e)
						{
							continue;
						}
						item.fromBytes(data);
						int version = item.getVersion();
						if(s == null || version != s.getVersion())
						{
							handler.handleError(str, item);
							map.client.delete().forPath(subPath+"/"+key);
						}
					}
					finally
					{
						lockManager.releaseLock(m);
					}
					
				}
			}
			
			Thread.sleep(SLEEP);
		}
	}
	
	private class Cleaner extends Thread
	{
		public void run()
		{
			while(true)
			{
				try
				{
					Thread.sleep(SLEEP);
					clean();
				}
				catch(Exception e)
				{
					//System.out.println("e");
					e.printStackTrace();
				}
			}
			
		}
	}
	
}
