package repository;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map.Entry;

import objectmap.KeyData;
import objectmap.ObjectGroup;
import objectmap.ObjectMap;
import remote.Byteable;

public class ParityFile implements Byteable
{	

	private int sizeGroup;
	private long gid;
	private byte data[];
	private HashMap<String, KeyData> keys;
	private int size;
	
	public static final int CHUNK_SIZE = 4096;
	
	
	public ParityFile()
	{
		size = 20;
	}
	
	public ParityFile(int sizeGroup, long groupID)
	{
		this.sizeGroup = sizeGroup;
		gid = groupID;
		data = new byte[ObjectMap.size[sizeGroup]*1024];
		keys = new HashMap<String,KeyData>();
		size = 20;
	}
	
	public String getName()
	{
		return ObjectGroup.keyFrom(gid, sizeGroup);
	}
	
	public DataReader degradedGet(String key, ObjectMap omap) throws Exception
	{
		if(keys.size() < 1)
		{
			return null;
		}
		int length = -1;
		for(Entry<String, KeyData> e : keys.entrySet())
		{
			if(e.getKey().equals(key))
			{
				length = e.getValue().size;
				continue;
			}
			int repo = e.getValue().repoID;
			String uploadString = e.getValue().getUniqueKey(e.getKey());
			Repository r = omap.repos.get(repo);
			InputStream data = r.get(uploadString, "");
			this.deleteError(data, e.getKey());
			data.close();
			
		}
		if(length == -1) return null;
		
		return new DataReader(new ByteArrayInputStream(this.data, 0, length),length);
	}
	
	public void update(String key, DataReader reader, KeyData kd) throws IOException
	{
		if(keys.put(key, kd) == null)
		{
			size += key.length();
		}
		int readerIndex = 0;
		int pfileReadIndex = 0;
		int readerRead;
		while( (readerRead = reader.update(readerIndex) ) != -1 )
		{
			int end = readerIndex + readerRead;
			while(readerIndex < end)
			{
				this.data[pfileReadIndex++] ^= reader.data[readerIndex++];
			}
		}
	}
	
	public int getCount()
	{
		return keys.size();
	}
	
	public void remove(String key, byte[] oldData)
	{
		if(keys.remove(key) == null) return;
		size -= key.length();
		
		for(int i = 0; i < oldData.length; i++)
		{	
			data[i] ^= oldData[i];	
		}
	}
	
	public ParityFile readFromStream(InputStream pfile) throws IOException
	{
		ObjectInputStream in = new ObjectInputStream(pfile);
		
		int n = in.readInt();
		this.keys = new HashMap<>(n);
		for(int i = 0; i < n; i++)
		{
			String s = in.readUTF();
			size += s.length();
			KeyData keyData = new KeyData();
			keyData.readFromStream(in);
			this.keys.put(s,keyData);
		}
		this.gid = in.readLong();
		this.sizeGroup = in.readInt();
		this.data = new byte[ObjectMap.size[this.sizeGroup]*1024];
		in.readFully(this.data, 0, this.data.length);
		return this;
	}

	public void readFromStream(InputStream pfile,
			String key, DataReader reader, KeyData kd) throws IOException
	{
		ObjectInputStream in = new ObjectInputStream(pfile);
		
		int n = in.readInt();
		this.keys = new HashMap<>(n);
		for(int i = 0; i < n; i++)
		{
			String s = in.readUTF();
			size += s.length();
			KeyData keyData = new KeyData();
			keyData.readFromStream(in);
			this.keys.put(s, keyData);
		}
		if(keys.put(key, kd) == null)
		{
			size += key.length();
		}
				
		this.gid = in.readLong();
		this.sizeGroup = in.readInt();
		this.data = new byte[ObjectMap.size[this.sizeGroup]*1024];
		
		
		int readerIndex = 0;
		int pfileReadIndex = 0;
		int pfileWriteIndex = 0;
		int readerRead;
		int pfileRead;
		while( (readerRead = reader.update(readerIndex) ) != -1 )
		{
			int end = readerIndex + readerRead;
			while(pfileWriteIndex < end)
			{
				int data = Math.min(this.data.length-pfileWriteIndex, CHUNK_SIZE);
				pfileRead = in.read(this.data, pfileWriteIndex, data);
				if(pfileRead == -1) throw new IOException("Parity File Done");
				pfileWriteIndex += pfileRead;
			}
			while(readerIndex < end)
			{
				this.data[pfileReadIndex++] ^= reader.data[readerIndex++];
			}
		}
		in.readFully(this.data, pfileWriteIndex, this.data.length - pfileWriteIndex);
	}
	
	public boolean deleteError(InputStream data, String key) throws IOException
	{
		byte[] buffer = new byte[ParityFile.CHUNK_SIZE];
		int read;
		int idx = 0;
		if(keys.remove(key) == null)
		{
			return false;
		}
		size -= key.length();
		while( (read = data.read(buffer, 0, ParityFile.CHUNK_SIZE)) != -1)
		{
			int r = 0;
			while(r < read)
			{
				this.data[idx++] ^= buffer[r++];
			}
		}
		return true;
	}
	
	public void deleteFromStream(InputStream pfile,
			String key, InputStream reader) throws IOException
	{
		ObjectInputStream in = new ObjectInputStream(pfile);
		
		int n = in.readInt();
		this.keys = new HashMap<>(n);
		for(int i = 0; i < n; i++)
		{
			String s = in.readUTF();
			size += s.length();
			KeyData keyData = new KeyData();
			keyData.readFromStream(in);
			this.keys.put(s, keyData);
		}
		if(keys.remove(key) != null)
		{
			size -= key.length();
		}
				
		this.gid = in.readLong();
		this.sizeGroup = in.readInt();
		this.data = new byte[ObjectMap.size[this.sizeGroup]*1024];
		
		
		int readerIndex = 0;
		int pfileReadIndex = 0;
		int pfileWriteIndex = 0;
		int readerRead;
		int pfileRead;
		byte[] buffer = new byte[ParityFile.CHUNK_SIZE];
		
		while( (readerRead = reader.read(buffer, 0, CHUNK_SIZE) ) != -1 )
		{
			int end = readerIndex + readerRead;
			while(pfileWriteIndex < end)
			{
				int data = Math.min(this.data.length-pfileWriteIndex, CHUNK_SIZE);
				pfileRead = in.read(this.data, pfileWriteIndex, data);
				if(pfileRead == -1) throw new IOException("Parity File Done");
				pfileWriteIndex += pfileRead;
			}
			int idx = 0;
			while(readerIndex < end)
			{
				readerIndex++;
				this.data[pfileReadIndex++] ^= buffer[idx++];
			}
		}
		in.readFully(this.data, pfileWriteIndex, this.data.length - pfileWriteIndex);
	}
	
	public void writeToStream(OutputStream stream) throws IOException
	{
		ObjectOutputStream out = new ObjectOutputStream(stream);
		out.writeInt(keys.size());
		for(Entry<String, KeyData> e : keys.entrySet())
		{
			out.writeUTF(e.getKey());
			KeyData re = e.getValue();
			re.writeToStream(out);
		}
		out.writeLong(gid);
		out.writeInt(sizeGroup);
		out.write(data,0,data.length);
		out.flush();
	}
	
	public ByteArrayInputStream getStream()
	{
		ByteArrayOutputStream os =
				new ByteArrayOutputStream(Math.max(this.data.length,this.data.length+4096));
		try {
			this.writeToStream(os);
		} catch (IOException e) {e.printStackTrace();}
		
		return new ByteArrayInputStream(os.toByteArray());
	}
	
	@Override
	public byte[] getBytes() {
		ByteArrayOutputStream os =
				new ByteArrayOutputStream(Math.max(this.data.length,this.data.length+4096));
		try {
			this.writeToStream(os);
		} catch (IOException e) {e.printStackTrace();}
		
		return os.toByteArray();
	}

	@Override
	public void fromBytes(byte[] bytes) {
		ByteArrayInputStream is = new ByteArrayInputStream(bytes);
		try {
			this.readFromStream(is);
		} catch (IOException e) {
		}
	}
}
